/*
 * fastpipe.c
 *
 * This is a simple one way communication construct between two threads or processes.  It is built on atomic operation,
 * the futex system call, and ipc shared memory.  Two instances could be combined to provide full duplex communication.
 *
 * It is designed to avoid intermediate copies and tries to efficiently use the buffer space.  The buffer space is treated as
 * a ring, and can be split up into multiple buffers.
 *
 *      Author: jhartmann
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <linux/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/syscall.h>
#include <sys/shm.h>
#include <unistd.h>
#include <linux/futex.h>
#include <sys/time.h>
#include <limits.h>
#include <time.h>
#include <error.h>
#include <errno.h>
#include "fastpipe.h"
#include "tiger.h"

#define DEBUG 0

int futex_wait(volatile int *uaddr, int val) {
	return syscall(__NR_futex, uaddr, FUTEX_WAIT, val, NULL, NULL, 0);
}

int futex_wake_all(volatile int *uaddr) {
	return syscall(__NR_futex, uaddr, FUTEX_WAKE, INT_MAX, NULL, NULL, 0);
}

fastpipe_buf_page *get_current_read_page(fastpipe *pipe) {
	int idx = atomic_read(&pipe->read_idx);
	return pipe->pages + idx;
}

fastpipe_buf_page *get_current_write_page(fastpipe *pipe) {
	int idx = atomic_read(&pipe->write_idx);
	return pipe->pages + idx;
}

void wake_readers(fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	int threshold = atomic_read(&pipe->read_threshold);
	chunk_list *used_list = handle->used_list;
	fill_used_list(handle);
	if(used_list->total_bytes >= threshold) {
		int value = cmpxchg(&pipe->read_futex, FASTPIPE_WAITING, FASTPIPE_WAKING);
		if(value == FASTPIPE_WAITING) {
			futex_wake_all(&pipe->read_futex);
		}
	}
}

void wake_writers(fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	int threshold = atomic_read(&pipe->write_threshold);
	chunk_list *free_list = handle->free_list;
	fill_free_list(handle);
	if(free_list->total_bytes >= threshold) {
		int value = cmpxchg(&pipe->write_futex, FASTPIPE_WAITING, FASTPIPE_WAKING);
		if(value == FASTPIPE_WAITING) {
			futex_wake_all(&pipe->write_futex);
		}
	}
}

void print_free_list(fastpipe_handle *handle, char *prefix) {
	chunk_list *list = handle->free_list;
	int i;
	printf("(%s) : Printing free list (total bytes: %d, num_chunks %d)\n", prefix, list->total_bytes, list->used_entries);
	for(i = 0; i < list->used_entries; i++) {
		chunk_entry *entry = list->entries + i;
		fastpipe_buf_page *page = handle->pipe->pages + entry->buffer_index;
		printf("(%s) : Entry (%d) buffer_idx: %d, entry_ofs: %d, addr: %p, size: %d\n", prefix, i, entry->buffer_index, entry->entry_ofs,
				entry->address, entry->num_bytes);
		printf("(%s) : Current read ofs: %d, Current write ofs: %d\n", prefix, atomic_read(&page->read_ofs), atomic_read(&page->write_ofs));
	}
}

void print_used_list(fastpipe_handle *handle, char *prefix) {
	chunk_list *list = handle->used_list;
	int i;
	printf("(%s) : Printing used list (total bytes: %d, num_chunks: %d)\n", prefix, list->total_bytes, list->used_entries);
	for(i = 0; i < list->used_entries; i++) {
		chunk_entry *entry = list->entries + i;
		fastpipe_buf_page *page = handle->pipe->pages + entry->buffer_index;
		printf("(%s) : Entry (%d) buffer_idx: %d, entry_ofs: %d, addr: %p, size: %d\n", prefix, i, entry->buffer_index, entry->entry_ofs, entry->address, entry->num_bytes);
		printf("(%s) : Current read ofs: %d, Current write ofs: %d\n", prefix, atomic_read(&page->read_ofs), atomic_read(&page->write_ofs));
	}
}

void wait_to_read(read_write_request *rqst, fastpipe_handle *handle) {
	if(DEBUG == 1) printf("Waiting to read\n");
	fastpipe *pipe = handle->pipe;
	chunk_list *list = handle->used_list;
	int value = 0;
	do {
		int slept = 0;
		value = cmpxchg(&pipe->read_futex, FASTPIPE_RUNNING, FASTPIPE_WAITING);
		if(value == FASTPIPE_RUNNING) {
			// We want to avoid a race here, we could have just gotten the lock right after a post from
			// a reader.  We want to make sure there is still no space in the buffer.
			fill_used_list(handle);
			if(list->total_bytes < rqst->minimum_threshold) {
				int err_val;
				int retval;
				slept = 1;
				atomic_inc(&pipe->read_wait_count);
				if(DEBUG == 1) {
					print_used_list(handle, "Reading before wait");
				}
				// blindly call wake on the write_futex
				futex_wake_all(&pipe->write_futex);
				retval = futex_wait(&pipe->read_futex, FASTPIPE_WAITING);
				err_val = errno;
				if(DEBUG == 1) printf("Reading: returning from futex wait, retval(%d), errno(%s)\n", retval, strerror(err_val));
			}
			value = xchg(&pipe->read_futex, FASTPIPE_RUNNING);
			// value of WAKING means we are waking up after the reader posted
			// io.  We might want to track this, but for now it doesn't matter
			if(slept == 1) {
				fill_used_list(handle);
				if(DEBUG == 1) print_used_list(handle, "Reading after_wait");
			}
		}
	} while(list->total_bytes < rqst->minimum_threshold);
}

void wait_to_write(read_write_request *rqst, fastpipe_handle *handle) {
	if(DEBUG == 1) printf("Waiting to write\n");
	fastpipe *pipe = handle->pipe;
	chunk_list *list = handle->free_list;
	int value = 0;
	int retval;
	do {
		int slept = 0;
		value = cmpxchg(&pipe->write_futex, FASTPIPE_RUNNING, FASTPIPE_WAITING);
		if(value == FASTPIPE_RUNNING) {
			// We want to avoid a race here, we could have just gotten the lock right after a post from
			// a reader.  We want to make sure there is still no space in the buffer.
			fill_free_list(handle);
			if(list->total_bytes < rqst->minimum_threshold) {
				int err_val;
				slept = 1;
				atomic_inc(&pipe->write_wait_count);
				if(DEBUG == 1) {
					print_free_list(handle, "Writing before wait");
				}
				// blindly call wake on read futex
				futex_wake_all(&pipe->read_futex);
				retval = futex_wait(&pipe->write_futex, FASTPIPE_WAITING);
				err_val = errno;
				if(DEBUG == 1) printf("Writing: returning from futex wait w/ retval(%d), errno(%s)\n", retval, strerror(err_val));
				if(DEBUG == 2) {
					printf("Sleeping for debug purposes");
					sleep(10000);
				}
			}
			value = xchg(&pipe->write_futex, FASTPIPE_RUNNING);
			// value of WAKING means we are waking up after the reader posted
			// io.  We might want to track this, but for now it doesn't matter
			if(slept == 1) {
				fill_free_list(handle);
				if(DEBUG == 1) {
					print_free_list(handle, "Writing after wait");
				}
			}
		}
	} while(list->total_bytes < rqst->minimum_threshold);
}

void request_read_io(read_write_request *rqst, fastpipe_handle *handle, int flags) {
	fastpipe *pipe = handle->pipe;
	atomic_set(&pipe->read_threshold, rqst->minimum_threshold);
	chunk_list *list = handle->used_list;
	rqst->chunks = list;
	rqst->bytes_processed = 0;
	if(list->total_bytes < rqst->minimum_threshold) {
		fill_used_list(handle);
		while(list->total_bytes < rqst->minimum_threshold && (flags & FASTPIPE_RQST_BLOCK) == FASTPIPE_RQST_BLOCK) {
			wait_to_read(rqst, handle);
			fill_used_list(handle);
		}
	}
}

void request_write_io(read_write_request *rqst, fastpipe_handle *handle, int flags) {
	// get the free list
	// if free list size > threshold just return the free list
	// if free list size < threshold, refresh the free list and see if it meets the condition
	// if blocking and < threshold sleep and then refresh the free list
	fastpipe *pipe = handle->pipe;
	atomic_set(&pipe->write_threshold, rqst->minimum_threshold);
	chunk_list *list = handle->free_list;
	rqst->chunks = list;
	rqst->bytes_processed = 0;
	if(list->total_bytes < rqst->minimum_threshold) {
		fill_free_list(handle);
		while(list->total_bytes < rqst->minimum_threshold && (flags & FASTPIPE_RQST_BLOCK) == FASTPIPE_RQST_BLOCK) {
			wait_to_write(rqst, handle);
			fill_free_list(handle);
		}
	}
}

void post_write_io(read_write_request *rqst, fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	chunk_list *list = handle->free_list;
	int remaining_bytes = rqst->bytes_processed;
	int i;
	for(i = 0; i < list->used_entries; i++) {
		chunk_entry *entry = list->entries + i;
		fastpipe_buf_page *page = pipe->pages + entry->buffer_index;
		atomic_set(&pipe->write_idx, entry->buffer_index);
		atomic_set(&page->write_ofs, entry->entry_ofs);
		int post_amount = (entry->num_bytes >= remaining_bytes ? remaining_bytes : entry->num_bytes);
		atomic_add(post_amount, &page->write_ofs);
		wake_readers(handle);
		remaining_bytes -= post_amount;
		if(remaining_bytes <= 0) {
			break;
		}
	}
	fill_free_list(handle);
	atomic_inc(&pipe->write_ops);
}

void post_read_io(read_write_request *rqst, fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	chunk_list *list = handle->used_list;
	int remaining_bytes = rqst->bytes_processed;
	int i;
	for(i = 0; i < list->used_entries; i++) {
		chunk_entry *entry = list->entries + i;
		fastpipe_buf_page *page = pipe->pages + entry->buffer_index;
		atomic_set(&pipe->read_idx, entry->buffer_index);
		atomic_set(&page->read_ofs, entry->entry_ofs);
		int post_amount = (entry->num_bytes >= remaining_bytes ? remaining_bytes : entry->num_bytes);
		atomic_add(post_amount, &page->read_ofs);
		wake_writers(handle);
		remaining_bytes -= post_amount;
		if(remaining_bytes <= 0) {
			break;
		}
	}
	fill_used_list(handle);
	atomic_inc(&pipe->read_ops);
}

fastpipe_handle *create_pipe(char *keyfile, int numBuffers, int bufferSize) {
	fastpipe_handle *handle = malloc(sizeof(fastpipe_handle));
	key_t shared_key = ftok(keyfile, 0xdeadbeef);
	int page_size = getpagesize();
	int shared_space_id = shmget(shared_key, (bufferSize * numBuffers) + page_size, IPC_CREAT | S_IRUSR | S_IWUSR | S_IROTH | S_IWOTH);
	if(shared_space_id == -1) {
		if(DEBUG == 1) {
			printf("create_pipe: Shared memory get failed\n");
		}
		return NULL;
	}
	u8 *buffer = (u8 *)shmat(shared_space_id, NULL, 0);
	handle->shm_root_address = (void *)buffer;
	handle->shm_id = shared_space_id;
	u8 *temp = buffer;
	int i;
	// The first page is spliced up for the fastpipe and
	// buffer structs
	fastpipe *pipe = (fastpipe *)buffer;
	memset((void *)pipe, 0, sizeof(fastpipe));
	temp += sizeof(fastpipe);
	memset((void *)temp, 0, sizeof(fastpipe_buf_page) * numBuffers);
	pipe->pages = (fastpipe_buf_page *)temp;
	buffer += page_size;
	for(i = 0; i < numBuffers; i++) {
		fastpipe_buf_page *page = pipe->pages + i;
		// The address must be an offset address
		page->buffer_idx = i;
		page->buffer_size = bufferSize;
		atomic_set(&page->read_ofs, 0);
		atomic_set(&page->write_ofs, 0);
	}
	pipe->read_futex = FASTPIPE_RUNNING;
	pipe->write_futex = FASTPIPE_RUNNING;
	atomic_set(&pipe->read_idx, 0);
	atomic_set(&pipe->write_idx, 0);
	// one byte default thresholds, api user can change.
	atomic_set(&pipe->read_threshold, 1);
	atomic_set(&pipe->write_threshold, 1);
	pipe->num_buffers = numBuffers;
	pipe->buffer_size = bufferSize;
	pipe->shared_key = shared_key;
	handle->pipe = pipe;
	handle->shared_address = buffer;
	handle->free_list = malloc(sizeof(chunk_list));
	handle->free_list->list_flags = CHUNK_LIST_FREE;
	handle->free_list->num_entries = 2 * numBuffers;
	handle->free_list->used_entries = 0;
	handle->free_list->entries = malloc(sizeof(chunk_entry) * handle->free_list->num_entries);
	memset((void *)handle->free_list->entries, 0, sizeof(chunk_entry) * handle->free_list->num_entries);
	handle->used_list = malloc(sizeof(chunk_list));
	handle->used_list->list_flags = CHUNK_LIST_USED;
	handle->used_list->num_entries = 2 * numBuffers;
	handle->used_list->used_entries = 0;
	handle->used_list->entries = malloc(sizeof(chunk_entry) * handle->used_list->num_entries);
	memset((void *)handle->used_list->entries, 0, sizeof(chunk_entry) * handle->used_list->num_entries);
	fill_free_list(handle);
	fill_used_list(handle);

	// statistics
	atomic_set(&pipe->write_wait_count, 0);
	atomic_set(&pipe->read_wait_count, 0);
	atomic_set(&pipe->read_ops, 0);
	atomic_set(&pipe->write_ops, 0);

	return handle;
}

// Called by user and creator to call shmdt on the shared buffer space.
void detach_pipe(fastpipe_handle *handle) {
	shmdt(handle->shm_root_address);
}

// Called by creator to detach
void cleanup_pipe(fastpipe_handle *handle) {
	struct shmid_ds shmid_struct;
    int result = shmctl(handle->shm_id, IPC_RMID, &shmid_struct);
    if(DEBUG == 1) {
    	printf("shmctl returned (%d)\n", result);
    }
}

// I might want to make incremental update functions as optimizations, but right now I will just fill the full
// free and used lists.
void fill_free_list(fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	int read_idx = atomic_read(&pipe->read_idx);
	int write_idx = atomic_read(&pipe->write_idx);
	int temp_write_idx = write_idx;
	int buffers_examined = 0;
	int buffer_size = pipe->buffer_size;
	int total_buffers = pipe->num_buffers;
	int total_bytes = 0;
	int num_chunks = 0;
	int next_buffer = 0;
	chunk_list *list = handle->free_list;
	list->used_entries = 0;
	do {
		fastpipe_buf_page *page = pipe->pages + temp_write_idx;
		u8 *page_addr = handle->shared_address + (temp_write_idx * buffer_size);
		int read_ofs = atomic_read(&page->read_ofs);
		int write_ofs = atomic_read(&page->write_ofs);
		int area_size = 0;
		if(write_ofs < read_ofs) {
			// just one chunk between the current write pointer and the read pointer.
			// we do not report the full size of this area, we just report its size - 1, or we don't report that
			// area at all if that result would be negative or zero.
			area_size = read_ofs - write_ofs;
			if(area_size > 0) area_size--;
			if(area_size > 0) {
				// Emit the chunk.
				chunk_entry *entry = list->entries + num_chunks;
				entry->address = page_addr + write_ofs;
				entry->entry_ofs = write_ofs;
				entry->buffer_index = temp_write_idx;
				entry->num_bytes = area_size;
				total_bytes += area_size;
				entry->readOfs = read_ofs;
				entry->writeOfs = write_ofs;
				num_chunks++;
			}
		} else {
			// There might be two chunks here.  If read_ofs != 0, then the free space is write_pointer to end of buffer &
			// top of buffer to current read pointer.
			if(read_ofs == 0 || read_ofs == 1) {
				// one chunk of free space.  report the whole amount
				area_size = buffer_size - write_ofs;
				if(area_size > 0) {
					// Emit the chunk.
					chunk_entry *entry = list->entries + num_chunks;
					entry->address = page_addr + write_ofs;
					entry->entry_ofs = write_ofs;
					entry->buffer_index = temp_write_idx;
					entry->num_bytes = area_size;
					total_bytes += area_size;
					entry->readOfs = read_ofs;
					entry->writeOfs = write_ofs;
					num_chunks++;
				}
			} else {
				// two chunks of free space.  report the whole amount on the first one, and size - 1 for the second one
				area_size = buffer_size - write_ofs;
				if(area_size > 0) {
					// Emit the chunks.
					chunk_entry *entry = list->entries + num_chunks;
					entry->address = page_addr + write_ofs;
					entry->buffer_index = temp_write_idx;
					entry->entry_ofs = write_ofs;
					entry->num_bytes = area_size;
					total_bytes += area_size;
					entry->readOfs = read_ofs;
					entry->writeOfs = write_ofs;
					num_chunks++;
				}
				// because of the test above for read_ofs == 0 or 1 we know there is at least one byte at the top so emit that chunk
				{
					chunk_entry *entry = list->entries + num_chunks;
					area_size = read_ofs - 1;
					entry->address = page_addr;
					entry->entry_ofs = 0;
					entry->buffer_index = temp_write_idx;
					entry->num_bytes = area_size;
					total_bytes += area_size;
					entry->readOfs = read_ofs;
					entry->writeOfs = write_ofs;
					num_chunks++;
				}
			}
		}
		temp_write_idx++;
		if(temp_write_idx >= total_buffers) temp_write_idx = 0;
		next_buffer = temp_write_idx;
		if(next_buffer >= total_buffers) next_buffer = 0;
		buffers_examined++;
	} while(buffers_examined < total_buffers && (next_buffer == read_idx || temp_write_idx == read_idx));
	// We want to stop when we are either one before the read_idx after wrapping around from our initial point, or on the read_idx
	// (will happen if there is just one buffer.)
	list->used_entries = num_chunks;
	list->total_bytes = total_bytes;
	list->list_flags = CHUNK_LIST_FREE;
}

void fill_used_list(fastpipe_handle *handle) {
	fastpipe *pipe = handle->pipe;
	int read_idx = atomic_read(&pipe->read_idx);
	int write_idx = atomic_read(&pipe->write_idx);
	int temp_read_idx = read_idx;
	int buffers_examined = 0;
	int buffer_size = pipe->buffer_size;
	int total_buffers = pipe->num_buffers;
	int total_bytes = 0;
	int num_chunks = 0;
	int prev_buffer = 0;
	chunk_list *list = handle->used_list;
	list->used_entries = 0;
	do {
		fastpipe_buf_page *page = pipe->pages + temp_read_idx;
		u8 *page_addr = handle->shared_address + (temp_read_idx * buffer_size);
		int read_ofs = atomic_read(&page->read_ofs);
		int write_ofs = atomic_read(&page->write_ofs);
		int area_size = 0;
		if(write_ofs < read_ofs) {
			// There are two chunks of data potentially here that are filled up.
			// We have the data from read_ofs to the end of the buffer, and the top
			// of the buffer to the write_ofs.
			area_size = buffer_size - read_ofs;
			if(area_size > 0) {
				// Emit the chunk.
				chunk_entry *entry = list->entries + num_chunks;
				entry->address = page_addr + read_ofs;
				entry->entry_ofs = read_ofs;
				entry->buffer_index = temp_read_idx;
				entry->num_bytes = area_size;
				total_bytes += area_size;
				entry->readOfs = read_ofs;
				entry->writeOfs = write_ofs;
				num_chunks++;
			}
			area_size = write_ofs;
			if(area_size > 0) {
				chunk_entry *entry = list->entries + num_chunks;
				entry->address = page_addr;
				entry->entry_ofs = 0;
				entry->buffer_index = temp_read_idx;
				entry->num_bytes += area_size;
				total_bytes += area_size;
				entry->readOfs = read_ofs;
				entry->writeOfs = write_ofs;
				num_chunks++;
			}
		} else {
			area_size = write_ofs - read_ofs;
			if(area_size > 0) {
				chunk_entry *entry = list->entries + num_chunks;
				entry->address = page_addr + read_ofs;
				entry->entry_ofs = read_ofs;
				entry->buffer_index = temp_read_idx;
				entry->num_bytes = area_size;
				total_bytes += area_size;
				entry->readOfs = read_ofs;
				entry->writeOfs = write_ofs;
				num_chunks++;
			}
		}
		prev_buffer = temp_read_idx;
		temp_read_idx++;
		if(temp_read_idx >= total_buffers) temp_read_idx = 0;
		buffers_examined++;
	} while(buffers_examined < total_buffers && prev_buffer != write_idx);
	// We want to stop when we just processed the buffer pointed at by the write idx.
	list->used_entries = num_chunks;
	list->total_bytes = total_bytes;
	list->list_flags = CHUNK_LIST_USED;
}

void test_write_blocks(fastpipe_handle *handle, int iterations, pid_t child) {
	printf("Starting write test\n");
	fastpipe *pipe = handle->pipe;
	long t1;
	long t2;
	int i, j;
	read_write_request *rqst = alloc_request(handle, FASTPIPE_RQST_TYPE_WRITE);
	double seconds;
	double rate;
	long long total_bytes = 0;
	char *page_aligned = valloc(4096);
	for(j = 0; j < 4096; j++) {
		page_aligned[j] = (j ^ 31) & 0xff;
	}
	rqst->minimum_threshold = 1024;
	t1 = clock();
	for(i=0; i < iterations; i++) {
		request_write_io(rqst, handle, FASTPIPE_RQST_BLOCK);
		chunk_list *chunks = rqst->chunks;
		chunk_entry *entry = chunks->entries;
		char *output = (char *)entry->address;
		int bytes_output = 0;
		if(entry->num_bytes >= 1024) {
			memcpy(output, page_aligned, 1024);
			bytes_output += 1024;
		} else {
			j = 0;
			while(bytes_output < 1024) {
				entry = chunks->entries + j;
				output = (char *)entry->address;
				int remaining = 1024 - bytes_output;
				int post_amt = entry->num_bytes < remaining ? entry->num_bytes : remaining;
				memcpy(output, page_aligned + bytes_output, post_amt);
				bytes_output += post_amt;
				j++;
			}
		}
		rqst->bytes_processed = bytes_output;
		total_bytes += bytes_output;
		post_write_io(rqst, handle);
	}
	t2 = clock();
	int status;
	waitpid(child, &status, 0);
	seconds = ((double)(t2 - t1)) / (double)CLOCKS_PER_SEC;
	rate = ((double)total_bytes / (double)(1024 * 1024)) / seconds;
	int write_ops = atomic_read(&pipe->write_ops);
	int read_ops = atomic_read(&pipe->read_ops);
	int write_waits = atomic_read(&pipe->write_wait_count);
	int read_waits = atomic_read(&pipe->read_wait_count);
	printf("Time 1: %ld, Time 2: %ld\n", t1, t2);
	printf("Number of seconds: %.4f\n", seconds);
	printf("Number of bytes: %lld\n", total_bytes);
	printf("Rate (MB/sec): %.4f\n", rate);
	printf("Number of read operations: %d\n", read_ops);
	printf("Number of write operations: %d\n", write_ops);
	printf("Number of blocking writes: %d\n", write_waits);
	printf("Number of blocking reads: %d\n", read_waits);
	printf("Write block ratio: %.4f\n", (double)write_waits / (double)write_ops);
	printf("Read block ratio: %.4f\n",  (double)read_waits / (double)read_ops);
	free_request(handle, rqst);
	detach_pipe(handle);
	free(page_aligned);
}

void test_write_blocks_validate(fastpipe_handle *handle, int iterations, pid_t child) {
	printf("Starting write validate test\n");
	fastpipe *pipe = handle->pipe;
	long t1;
	long t2;
	int i, j;
	read_write_request *rqst = alloc_request(handle, FASTPIPE_RQST_TYPE_WRITE);
	double seconds;
	double rate;
	long long total_bytes = 0;
	char *page_aligned = valloc(4096);
	srandom(1);

	for(j = 0; j < 4096; j++) {
		page_aligned[j] = (j ^ 31) & 0xff;
	}
	// Put the last 3 64 bit words as the tiger hash of the 4 pages.
	for(j = 0; j < 4; j++) {
		int ofs = j * 1024;
		int length = 1024 - (sizeof(word64) * 3);
		word64* result = (word64 *)(page_aligned + (ofs + length));
		tiger((word64 *)(page_aligned + ofs), length, result);
	}
	rqst->minimum_threshold = 1024;
	t1 = clock();
	for(i=0; i < iterations; i++) {
		request_write_io(rqst, handle, FASTPIPE_RQST_BLOCK);
		chunk_list *chunks = rqst->chunks;
		chunk_entry *entry = chunks->entries;
		char *output = (char *)entry->address;
		int bytes_output = 0;
		long random_val = random();
		int ofs = (random_val % 4) * 1024;
		char *input = page_aligned + ofs;
		if(entry->num_bytes >= 1024) {
			memcpy(output, input, 1024);
			bytes_output += 1024;
		} else {
			j = 0;
			while(bytes_output < 1024) {
				entry = chunks->entries + j;
				output = (char *)entry->address;
				int remaining = 1024 - bytes_output;
				int post_amt = entry->num_bytes < remaining ? entry->num_bytes : remaining;
				memcpy(output, input + bytes_output, post_amt);
				bytes_output += post_amt;
				j++;
			}
		}
		rqst->bytes_processed = bytes_output;
		total_bytes += bytes_output;
		post_write_io(rqst, handle);
	}
	t2 = clock();
	int status;
	waitpid(child, &status, 0);
	seconds = ((double)(t2 - t1)) / (double)CLOCKS_PER_SEC;
	rate = ((double)total_bytes / (double)(1024 * 1024)) / seconds;
	int write_ops = atomic_read(&pipe->write_ops);
	int read_ops = atomic_read(&pipe->read_ops);
	int write_waits = atomic_read(&pipe->write_wait_count);
	int read_waits = atomic_read(&pipe->read_wait_count);
	printf("Time 1: %ld, Time 2: %ld\n", t1, t2);
	printf("Number of seconds: %.4f\n", seconds);
	printf("Number of bytes: %lld\n", total_bytes);
	printf("Rate (MB/sec): %.4f\n", rate);
	printf("Number of read operations: %d\n", read_ops);
	printf("Number of write operations: %d\n", write_ops);
	printf("Number of blocking writes: %d\n", write_waits);
	printf("Number of blocking reads: %d\n", read_waits);
	printf("Write block ratio: %.4f\n", (double)write_waits / (double)write_ops);
	printf("Read block ratio: %.4f\n",  (double)read_waits / (double)read_ops);
	free_request(handle, rqst);
	detach_pipe(handle);
	free(page_aligned);
}

/* simple validation read, the write process is writing out 1k blocks where the last 3 64-bit words are the tiger hash
 * of the other bit.  This tests if they are equal. */
void test_read_validate(fastpipe_handle *handle, int iterations) {
	printf("Starting read validate test\n");
	fastpipe *pipe = handle->pipe;
	read_write_request *rqst = alloc_request(handle, FASTPIPE_RQST_TYPE_READ);
	char *page_aligned = valloc(4096);
	word64 result[3];
	word64 *result_ptr;
	int length = 1024 - (sizeof(word64) * 3);
	int i, j;
	int bytes_read = 0;
	long long matches = 0;
	long long failures = 0;
	rqst->minimum_threshold = 1024;
	for(i = 0; i < iterations;i++) {
		request_read_io(rqst, handle, FASTPIPE_RQST_BLOCK);
		chunk_list *chunks = rqst->chunks;
		chunk_entry *entry = chunks->entries;
		char *input = (char *)entry->address;
		bytes_read = 0;

		if(entry->num_bytes >= 1024) {
			tiger((word64 *)input, length, result);
			input = (char *)entry->address + length;
			result_ptr = (word64 *)input;
			bytes_read += 1024;
		} else {
			j = 0;
			while(bytes_read < 1024) {
				entry = chunks->entries + j;
				input = (char *)entry->address;
				int remaining = 1024 - bytes_read;
				int post_amt = entry->num_bytes < remaining ? entry->num_bytes : remaining;
				memcpy(page_aligned + bytes_read, input, post_amt);
				bytes_read += post_amt;
				j++;
			}
			tiger((word64 *)page_aligned, length, result);
			result_ptr = (word64 *)(page_aligned + length);
		}
		j = memcmp(result_ptr, result, sizeof(word64) * 3);
		if(j == 0) {
			matches++;
		} else {
			failures++;
		}
		rqst->bytes_processed = bytes_read;
		post_read_io(rqst, handle);
	}
	int read_ops = atomic_read(&pipe->read_ops);
	double success_ratio = (double)matches / (double)read_ops;
	printf("Matches: %lld, failures: %lld, success ratio: %.4f\n", matches, failures, success_ratio);
	free_request(handle, rqst);
	detach_pipe(handle);
	free(page_aligned);
}

 /* simple read code that just reads, or skips the input */
void test_read(fastpipe_handle *handle, int iterations, int actually_read) {
	printf("Starting read test\n");
	read_write_request *rqst = alloc_request(handle, FASTPIPE_RQST_TYPE_READ);
	char *page_aligned = valloc(4096);
	chunk_list *used_list = handle->used_list;
	int bytes_read = 0;
	int i, j;
	rqst->minimum_threshold = 1024;
	for(i = 0; i < iterations;i++) {
		request_read_io(rqst, handle, FASTPIPE_RQST_BLOCK);
		if(actually_read == 1) {
			chunk_list *chunks = rqst->chunks;
			chunk_entry *entry = chunks->entries;
			char *input = (char *)entry->address;
			bytes_read = 0;

			if(entry->num_bytes >= 1024) {
				memcpy(page_aligned, input, 1024);
				bytes_read += 1024;
			} else {
				j = 0;
				while(bytes_read < 1024) {
					entry = chunks->entries + j;
					input = (char *)entry->address;
					int remaining = 1024 - bytes_read;
					int post_amt = entry->num_bytes < remaining ? entry->num_bytes : remaining;
					memcpy(page_aligned + bytes_read, input, post_amt);
					bytes_read += post_amt;
					j++;
				}
			}
			rqst->bytes_processed = bytes_read;
		} else {
			rqst->bytes_processed = (used_list->total_bytes > 1024 ? 1024 : used_list->total_bytes);
		}
		post_read_io(rqst, handle);
	}
	free_request(handle, rqst);
	detach_pipe(handle);
	free(page_aligned);
}

read_write_request *alloc_request(fastpipe_handle *handle, int flags) {
	read_write_request *rqst = malloc(sizeof(read_write_request));
	rqst->bytes_processed = 0;
	rqst->minimum_threshold = 1;
	if((flags & FASTPIPE_RQST_TYPE_WRITE) == FASTPIPE_RQST_TYPE_WRITE) {
		rqst->chunks = handle->free_list;
	} else {
		rqst->chunks = handle->used_list;
	}
	return rqst;
}

void free_request(fastpipe_handle *handle, read_write_request *rqst) {
	free(rqst);
}

int main(int argc, char **argv) {
	//char *keyfile = "/tmp/test.dat";
	char *keyfile = "/dev/null";
	int num_buffers = 1;
	int buffer_size_in_pages = 16;
	int iterations = 10000000;
	int validate_iter = 2000000;
	int page_size = getpagesize();
	fastpipe_handle *handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    pid_t child;
    printf("Beginning test : write only, read is a noop (one 64k buffer).\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 0);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : read and write. (one 64k buffer)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 1);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : data that is written is validated as it is read. (one 64k buffer)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read_validate(handle, validate_iter);
    		exit(0);
    	} else {
    		test_write_blocks_validate(handle, validate_iter, child);
    	}
    }
    cleanup_pipe(handle);
    num_buffers = 4;
    buffer_size_in_pages = 4;
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : write only, read is a noop (four 16k buffers).\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 0);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : read and write. (four 16k buffers)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 1);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : data that is written is validated as it is read. (four 16k buffers)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read_validate(handle, validate_iter);
    		exit(0);
    	} else {
    		test_write_blocks_validate(handle, validate_iter, child);
    	}
    }
    cleanup_pipe(handle);

    num_buffers = 8;
    buffer_size_in_pages = 1;
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : write only, read is a noop (eight 4k buffers).\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 0);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : read and write. (eight 4k buffers)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read(handle, iterations, 1);
    		exit(0);
    	} else {
    		test_write_blocks(handle, iterations, child);
    	}
    }
    cleanup_pipe(handle);
	handle = create_pipe(keyfile, num_buffers, page_size * buffer_size_in_pages);
    printf("Beginning test : data that is written is validated as it is read. (eight 4k buffers)\n");
    child = fork();
    if(child >= 0) {
    	if(child == 0) {
    		test_read_validate(handle, validate_iter);
    		exit(0);
    	} else {
    		test_write_blocks_validate(handle, validate_iter, child);
    	}
    }
    cleanup_pipe(handle);

	return 0;
}
