/*
 * fastpipe.h
 *
 * Lots of functions here are part of the linux kernel source and are licensed under the GPL.
 *
 *      Author: jhartmann
 */

#ifndef FASTPIPE_H_
#define FASTPIPE_H_

typedef unsigned char u8;

typedef struct {
	volatile int counter;
} atomic_t;

#ifdef __ASSEMBLY__
# define __ASM_FORM(x)	x
# define __ASM_EX_SEC	.section __ex_table, "a"
#else
# define __ASM_FORM(x)	" " #x " "
# define __ASM_EX_SEC	" .section __ex_table,\"a\"\n"
#endif

# define __ASM_SEL(a,b) __ASM_FORM(a)
/*
#ifdef CONFIG_X86_32
# define __ASM_SEL(a,b) __ASM_FORM(a)
#else
# define __ASM_SEL(a,b) __ASM_FORM(b)
#endif
*/
#define __ASM_SIZE(inst)	__ASM_SEL(inst##l, inst##q)
#define __ASM_REG(reg)		__ASM_SEL(e##reg, r##reg)

#define _ASM_PTR	__ASM_SEL(.long, .quad)
#define _ASM_ALIGN	__ASM_SEL(.balign 4, .balign 8)

#define LOCK_PREFIX \
		".section .smp_locks,\"a\"\n"	\
		_ASM_ALIGN "\n"			\
		_ASM_PTR "661f\n" /* address */	\
		".previous\n"			\
		"661:\n\tlock; "

#define mb() 	asm volatile("mfence":::"memory")
#define rmb()	asm volatile("lfence":::"memory")
#define wmb()	asm volatile("sfence" ::: "memory")

struct __xchg_dummy {
	unsigned long a[100];
};
#define __xg(x) ((struct __xchg_dummy *)(x))

#define cmpxchg(ptr, o, n)						\
	((__typeof__(*(ptr)))__cmpxchg((ptr), (unsigned long)(o),	\
				       (unsigned long)(n),		\
				       sizeof(*(ptr))))

static inline unsigned long __cmpxchg(volatile void *ptr, unsigned long old,
				      unsigned long new, int size)
{
	unsigned long prev;
	switch (size) {
	case 1:
		asm volatile(LOCK_PREFIX "cmpxchgb %b1,%2"
			     : "=a"(prev)
			     : "q"(new), "m"(*__xg(ptr)), "0"(old)
			     : "memory");
		return prev;
	case 2:
		asm volatile(LOCK_PREFIX "cmpxchgw %w1,%2"
			     : "=a"(prev)
			     : "r"(new), "m"(*__xg(ptr)), "0"(old)
			     : "memory");
		return prev;
	case 4:
		asm volatile(LOCK_PREFIX "cmpxchgl %1,%2"
			     : "=a"(prev)
			     : "r"(new), "m"(*__xg(ptr)), "0"(old)
			     : "memory");
		return prev;
	}
	return old;
}

/**
 * atomic_inc - increment atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically increments @v by 1.
 */
static inline void atomic_inc(atomic_t *v)
{
	asm volatile(LOCK_PREFIX "incl %0"
		     : "+m" (v->counter));
}

/**
 * atomic_read - read atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically reads the value of @v.
 */
static inline int atomic_read(const atomic_t *v)
{
	return v->counter;
}

/**
 * atomic_set - set atomic variable
 * @v: pointer of type atomic_t
 * @i: required value
 *
 * Atomically sets the value of @v to @i.
 */
static inline void atomic_set(atomic_t *v, int i)
{
	v->counter = i;
}

/**
 * atomic_add - add integer to atomic variable
 * @i: integer value to add
 * @v: pointer of type atomic_t
 *
 * Atomically adds @i to @v.
 */
static inline void atomic_add(int i, atomic_t *v)
{
	asm volatile(LOCK_PREFIX "addl %1,%0"
		     : "+m" (v->counter)
		     : "ir" (i));
}

/**
 * atomic_sub - subtract integer from atomic variable
 * @i: integer value to subtract
 * @v: pointer of type atomic_t
 *
 * Atomically subtracts @i from @v.
 */
static inline void atomic_sub(int i, atomic_t *v)
{
	asm volatile(LOCK_PREFIX "subl %1,%0"
		     : "+m" (v->counter)
		     : "ir" (i));
}

/**
 * atomic_add_return - add integer and return
 * @v: pointer of type atomic_t
 * @i: integer value to add
 *
 * Atomically adds @i to @v and returns @i + @v
 */
static inline int atomic_add_return(int i, atomic_t *v)
{
	int __i;
	__i = i;
	asm volatile(LOCK_PREFIX "xaddl %0, %1"
		     : "+r" (i), "+m" (v->counter)
		     : : "memory");
	return i + __i;
}

/**
 * atomic_sub_return - subtract integer and return
 * @v: pointer of type atomic_t
 * @i: integer value to subtract
 *
 * Atomically subtracts @i from @v and returns @v - @i
 */
static inline int atomic_sub_return(int i, atomic_t *v)
{
	return atomic_add_return(-i, v);
}

#define xchg(ptr, v)							\
	((__typeof__(*(ptr)))__xchg((unsigned long)(v), (ptr), sizeof(*(ptr))))

/*
 * Note: no "lock" prefix even on SMP: xchg always implies lock anyway
 * Note 2: xchg has side effect, so that attribute volatile is necessary,
 *	  but generally the primitive is invalid, *ptr is output argument. --ANK
 */
static inline unsigned long __xchg(unsigned long x, volatile void *ptr,
				   int size)
{
	switch (size) {
	case 1:
		asm volatile("xchgb %b0,%1"
			     : "=q" (x)
			     : "m" (*__xg(ptr)), "0" (x)
			     : "memory");
		break;
	case 2:
		asm volatile("xchgw %w0,%1"
			     : "=r" (x)
			     : "m" (*__xg(ptr)), "0" (x)
			     : "memory");
		break;
	case 4:
		asm volatile("xchgl %0,%1"
			     : "=r" (x)
			     : "m" (*__xg(ptr)), "0" (x)
			     : "memory");
		break;
	}
	return x;
}

#define CHUNK_LIST_USED 0x1
#define CHUNK_LIST_FREE 0x2

typedef struct _chunk_entry {
	u8 *address;
	int num_bytes;
	int buffer_index;
	int entry_ofs;
	// This is the read and write offset from the shared area.  We copy them to from the shared area
	// as we walk it to build the list.  This allows us to check for stale lists with a quick operation
	// later.  We don't want people using the api to use it directly.
	int readOfs;
	int writeOfs;
} chunk_entry;

// There are two sets of chunk lists per direction on the pipe.  They are not in shared memory, they are local
// structures that are built from the shared data as needed.  One is the chunks of data on the ring, the other
// is the chunks of free space available for writing.
typedef struct _chunk_list {
	int list_flags; /* If CHUNK_LIST_USED is set, then the chunk list describes the areas that are readable in the buffer set.
					   If CHUNK_LIST_FREE is set, then the chunk list describes the areas that are writable in the buffer set. */
	int total_bytes;
	// There will always be 2 * number of buffers records in this list, since that is the maximum
	// amount of items that are required to describe the free space in the buffer set.
	int num_entries;
	// the number of used entries
	int used_entries;
	chunk_entry* entries; /* always allocated at the maximum size, which is number of buffers * 2.  We only use exactly what is needed to
							 describe the free / used space */
} chunk_list;

// We want to do atomic adds / xchg to
// the read and write pointers.  We only want
// to sleep if the queue is completely full or
// completely empty
typedef struct _fastpipe_buf_page {
	atomic_t read_ofs;
	atomic_t write_ofs;
	int buffer_idx;
	int buffer_size;
} fastpipe_buf_page;

typedef struct _fastpipe {
	int buffer_size;
	int num_buffers;
	volatile int read_futex;
	volatile int write_futex;
	atomic_t read_idx;
	atomic_t write_idx;
	atomic_t read_threshold;
	atomic_t write_threshold;
	key_t shared_key;
	atomic_t write_wait_count;
	atomic_t read_wait_count;
	atomic_t write_ops;
	atomic_t read_ops;

	fastpipe_buf_page *pages;
} fastpipe;

typedef struct _fastpipe_handle {
	u8 *shared_address;
	fastpipe *pipe;
	chunk_list *free_list;
	chunk_list *used_list;
	void *shm_root_address;
	int shm_id;
} fastpipe_handle;

#define FASTPIPE_WAITING 0x1
#define FASTPIPE_RUNNING 0x2
#define FASTPIPE_WAKING  0x3

/* Flags to pass into the read/write request api */
#define FASTPIPE_RQST_BLOCK	0x1

/* Request type to pass into alloc_request */
#define FASTPIPE_RQST_TYPE_READ 0x1
#define FASTPIPE_RQST_TYPE_WRITE 0x2

/* Always assume linear usage of memory, since that is what is required by the api contract.  Because of that we only need
 * one variable to describe the space consumed from the chunk list. */
typedef struct _read_write_request {
	// The minimum threshold required for the thread to wake up if waiting, or the condition to sleep on.
	int minimum_threshold;
	int bytes_processed;
	// This just points at the master chunk list
	chunk_list *chunks;
} read_write_request;

/* internal functions - start */
void fill_free_list(fastpipe_handle *handle);
void fill_used_list(fastpipe_handle *handle);

int futex_wait(volatile int *uaddr, int val);
int futex_wake_all(volatile int *uaddr);

void wake_readers(fastpipe_handle *handle);
void wake_writers(fastpipe_handle *handle);

void wait_to_read(read_write_request *rqst, fastpipe_handle *handle);
void wait_to_write(read_write_request *rqst, fastpipe_handle *handle);

/* internal functions - end */

/* public interface */

/*
 * Allocate a request object to use with request_read_io, request_write_io, post_read_io, & post_write_io
 */
read_write_request *alloc_request(fastpipe_handle *handle, int flags);
/*
 * free a request object returned from alloc_request.
 */
void free_request(fastpipe_handle *handle, read_write_request *rqst);

/*
 * Request a pointer and count of bytes available to be read from the
 * returned pointer.  The caller doesn't have to worry about the ring buffer
 * nature of the implementation.  It just processes the requests without concern
 * for wrapping.
 * The caller can pass in flags with the FASTPIPE_RQST_BLOCK bit set, and the
 * request will block if until there is something to be read.
 */
void request_read_io(read_write_request *rqst, fastpipe_handle *handle, int flags);

/*
 * Request a pointer and count of the space available to be written to in the
 * returned pointer.  The caller doesn't have to worry about the ring buffer
 * nature of the implementation.
 * The caller can pass in flags with the FASTPIPE_RQST_BLOCK bit set, and the
 * request will block if until there is room for something to be written.
 *
 */
void request_write_io(read_write_request *rqst, fastpipe_handle *handle, int flags);

/*
 * Post the number of bytes written to the pipe.  The read_write_request must contain the
 * number of bytes written to the provided address.  It must not be more then the reported capacity.
 */
void post_write_io(read_write_request *rqst, fastpipe_handle *handle);

/*
 * Post the number of bytes read from the pipe.  The read_write_request must contain the number
 * of bytes read from the provided address.  It must not be more then the reported available bytes.
 */
void post_read_io(read_write_request *rqst, fastpipe_handle *handle);

fastpipe_handle *create_pipe(char *keyfile, int numBuffers, int bufferSize);
// Called by user and creator to call shmdt on the shared buffer space.
void detach_pipe(fastpipe_handle *handle);
// Called by creator to detach
void cleanup_pipe(fastpipe_handle *handle);
#endif /* FASTPIPE_H_ */
